from django.db.models.signals import pre_delete
from django.apps import apps
from django.core.checks import Error, register
from .models import Domain
from .providers import get_domain_provider


def remove_domain(sender, instance: Domain, **kwargs):
    provider = get_domain_provider(instance.provider)
    if provider:
        provider.remove_domain(instance)


pre_delete.connect(remove_domain, sender=Domain)


def register_checks():
    def check(app_configs, **kwargs):
        if not apps.is_installed('saas_base'):
            return [Error("'saas_base' must be in INSTALLED_APPS.")]
        else:
            return []

    register(check, 'saas')



from abc import ABCMeta, abstractmethod
from ..models import Domain


class BaseProvider(object, metaclass=ABCMeta):
    def __init__(self, **options):
        self.options = options

    @abstractmethod
    def add_domain(self, domain: Domain) -> Domain:
        ...

    @abstractmethod
    def verify_domain(self, domain: Domain) -> Domain:
        ...

    @abstractmethod
    def remove_domain(self, domain: Domain):
        ...



from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from .settings import domain_settings
from .models import Domain
from .signals import before_add_domain, after_add_domain


class DomainSerializer(serializers.ModelSerializer):
    class Meta:
        model = Domain
        exclude = ["tenant", "instrument_id"]

    def validate_provider(self, value: str):
        if value in domain_settings.PROVIDERS:
            return value
        raise ValidationError(f"Provider '{value}' is not supported")

    def create(self, validated_data):
        before_add_domain.send(self.__class__, data=validated_data, **self.context)
        instance = super().create(validated_data)
        after_add_domain.send(self.__class__, instance=instance, **self.context)
        return instance



import logging
from .base import BaseProvider
from ..models import Domain

logger = logging.getLogger(__name__)


class NullProvider(BaseProvider):
    def add_domain(self, domain: Domain) -> Domain:
        logger.info("Enable domain:", domain)
        domain.verified = True
        domain.ssl = self.options.get("ssl", False)
        domain.save()
        return domain

    def verify_domain(self, domain: Domain) -> Domain:
        logger.info("Verify domain:", domain)
        return domain

    def remove_domain(self, domain: Domain):
        logger.info("Remove domain:", domain)



from .base import BaseProvider
from .null import NullProvider
from .cloudflare import CloudflareProvider
from ..settings import domain_settings


def get_domain_provider(name: str):
    return domain_settings.PROVIDERS.get(name)


__all__ = [
    "get_domain_provider",
    "BaseProvider",
    "NullProvider",
    "CloudflareProvider",
]



import uuid
from django.conf import settings
from django.db import migrations, models
from django.utils import timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.SAAS_TENANT_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Domain',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('provider', models.CharField(max_length=100)),
                ('hostname', models.CharField(max_length=100, unique=True)),
                ('verified', models.BooleanField(default=False, editable=False)),
                ('ssl', models.BooleanField(default=False, editable=False)),
                ('active', models.BooleanField(default=False, editable=False)),
                ('instrument_id', models.CharField(blank=True, editable=False, max_length=256, null=True)),
                ('instrument', models.JSONField(blank=True, editable=False, null=True)),
                ('created_at', models.DateTimeField(default=timezone.now, editable=False)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('tenant', models.ForeignKey(on_delete=models.CASCADE, to=settings.SAAS_TENANT_MODEL)),
            ],
            options={
                'verbose_name': 'domain',
                'verbose_name_plural': 'domains',
                'db_table': 'saas_domain',
                'ordering': ['created_at'],
            },
        ),
    ]



from django.urls import path
from .endpoints import DomainListEndpoint, DomainItemEndpoint

urlpatterns = [
    path('domains', DomainListEndpoint.as_view()),
    path('domains/<pk>', DomainItemEndpoint.as_view()),
]



from django.core.signals import setting_changed
from saas_base.settings import Settings

DEFAULTS = {'PROVIDERS': {}}


class DomainSettings(Settings):
    IMPORT_PROVIDERS = [
        'PROVIDERS',
    ]

domain_settings = DomainSettings('SAAS_DOMAIN', defaults=DEFAULTS)
setting_changed.connect(domain_settings.listen_setting_changed)



from django.dispatch import Signal

before_add_domain = Signal()
after_add_domain = Signal()
before_remove_domain = Signal()

__all__ = [
    "before_add_domain",
    "after_add_domain",
    "before_remove_domain",
]



from rest_framework.mixins import ListModelMixin
from rest_framework.request import Request
from rest_framework.response import Response
from saas_base.drf.views import TenantEndpoint
from ..serializers import DomainSerializer
from ..providers import get_domain_provider
from ..models import Domain


class DomainListEndpoint(ListModelMixin, TenantEndpoint):
    serializer_class = DomainSerializer
    pagination_class = None
    queryset = Domain.objects.all()
    resource_name = 'tenant'
    resource_http_method_actions = {
        'GET': 'read',
        'POST': 'admin',
    }

    def get(self, request: Request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request: Request, *args, **kwargs):
        tenant_id = self.get_tenant_id()
        serializer: DomainSerializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        domain = serializer.save(tenant_id=tenant_id)
        provider = get_domain_provider(domain.provider)
        if provider:
            provider.add_domain(domain)
            serializer = self.get_serializer(domain)
        return Response(serializer.data, status=201)


class DomainItemEndpoint(TenantEndpoint):
    serializer_class = DomainSerializer
    queryset = Domain.objects.all()
    resource_name = 'tenant'
    resource_http_method_actions = {
        'GET': 'read',
        'POST': 'admin',
        'DELETE': 'admin',
    }

    def get(self, request: Request, *args, **kwargs):
        domain = self.get_object()
        if request.query_params.get('verify'):
            provider = get_domain_provider(domain.provider)
            if provider:
                provider.verify_domain(domain)
        serializer: DomainSerializer = self.get_serializer(domain)
        return Response(serializer.data)

    def post(self, request: Request, *args, **kwargs):
        domain = self.get_object()
        provider = get_domain_provider(domain.provider)
        if provider:
            provider.add_domain(domain)
        serializer: DomainSerializer = self.get_serializer(domain)
        return Response(serializer.data)

    def delete(self, request: Request, *args, **kwargs):
        domain = self.get_object()
        provider = get_domain_provider(domain.provider)
        if provider:
            provider.remove_domain(domain)
        domain.delete()
        return Response('', status=204)



from django.http import HttpResponse
from django.test import RequestFactory
from saas_domain.middleware import DomainTenantIdMiddleware
from saas_domain.models import Domain
from saas_base.test import SaasTestCase


class TestDomainTenantIdMiddleware(SaasTestCase):
    middleware_cls = DomainTenantIdMiddleware

    @staticmethod
    def create_request(host):
        meta = {'HTTP_HOST': host}
        factory = RequestFactory(**meta)
        return factory.get('/')

    def setUp(self):
        self.middleware = self.middleware_cls(lambda req: HttpResponse())

    def test_match_tenant_id(self):
        Domain.objects.create(hostname='test.us.localhost', tenant=self.tenant)
        request = self.create_request('test.us.localhost')
        self.middleware(request)
        self.assertEqual(request.tenant_id, self.tenant.pk)

    def test_none_tenant_id(self):
        request = self.create_request('none.us.localhost')
        self.middleware(request)
        self.assertEqual(request.tenant_id, None)

    def test_cached_tenant_id(self):
        request = self.create_request('test.us.localhost')
        request._cached_tenant_id = self.tenant.pk
        self.middleware(request)
        self.assertEqual(request.tenant_id, self.tenant.pk)
